from __future__ import annotations

import re
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Dict,
    Generator,
    Iterable,
    Literal,
    TypeVar,
    overload,
)

from typing_extensions import Concatenate, ParamSpec

from .guard import is_async

T = TypeVar("T")


def group_dict(iterable: Iterable, key_callable: Callable[[Any], Any]):
    temp = {}
    for i in iterable:
        k = key_callable(i)
        temp.setdefault(k, []).append(i)
    return temp


async def run_always_await(target: Callable[..., Any] | Callable[..., Awaitable[Any]], *args, **kwargs) -> Any:
    obj = target(*args, **kwargs)
    if is_async(target) or is_async(obj):
        obj = await obj  # type: ignore
    return obj


def gen_subclass(cls: type[T]) -> Generator[type[T], None, None]:
    """生成某个类的所有子类 (包括其自身)
    Args:
        cls (Type[T]): 类
    Yields:
        Type[T]: 子类
    """
    yield cls
    for sub_cls in cls.__subclasses__():
        if TYPE_CHECKING:
            assert issubclass(sub_cls, cls)
        yield from gen_subclass(sub_cls)


R = TypeVar("R")
P = ParamSpec("P")


@overload
def init_spec(fn: Callable[P, T]) -> Callable[[Callable[[T], R]], Callable[P, R]]: ...


@overload
def init_spec(
    fn: Callable[P, T], is_method: Literal[True]
) -> Callable[[Callable[[Any, T], R]], Callable[Concatenate[Any, P], R]]: ...


def init_spec(  # type: ignore
    fn: Callable[P, T], is_method: bool = False
) -> Callable[[Callable[[T], R] | Callable[[Any, T], R]], Callable[P, R] | Callable[Concatenate[Any, P], R]]:
    def wrapper(func: Callable[[T], R] | Callable[[Any, T], R]) -> Callable[P, R] | Callable[Concatenate[Any, P], R]:
        def inner(*args: P.args, **kwargs: P.kwargs):
            if is_method:
                return func(args[0], fn(*args[1:], **kwargs))  # type: ignore
            return func(fn(*args, **kwargs))  # type: ignore

        return inner

    return wrapper


def safe_eval(route: str, _locals: Dict[str, Any]):
    parts = re.split(r"\.|(\[.+?\])|(\(.*?\))", route)
    if (key := parts[0]) not in _locals:
        raise NameError(key)
    res = _locals[key]
    for part in parts[1:]:
        if not part:
            continue
        if part.startswith("_"):
            raise ValueError(route)
        if part.startswith("[") and part.endswith("]"):
            item = part[1:-1]
            if item[0] in ("'", '"') and item[-1] in ("'", '"'):
                res = res[item[1:-1]]
            elif ":" in item:
                res = res[slice(*(int(x) if x else None for x in item.split(":")))]
            else:
                try:
                    res = res[int(item)]
                except ValueError:
                    res = res[item]
        elif part.startswith("(") and part.endswith(")"):
            item = part[1:-1]
            if not item:
                res = res()
            else:
                _parts = item.split(",")
                _args = []
                _kwargs = {}
                for _part in _parts:
                    _part = _part.strip()
                    if re.match(".+=.+", _part):
                        k, v = _part.split("=")
                        if v[0] in ("'", '"') and v[-1] in ("'", '"'):
                            v = v[1:-1]
                        _kwargs[k] = v
                    else:
                        if _part[0] in ("'", '"') and _part[-1] in ("'", '"'):
                            _part = _part[1:-1]
                        _args.append(_part)
                res = res(*_args, **_kwargs)
        else:
            res = getattr(res, part)
    return res


def uncapitalize(source: str) -> str:
    return source[0].lower() + source[1:]


def camel_case(source: str) -> str:
    return re.sub("[_-][a-z]", lambda mat: mat[0][1:].upper(), source)


def pascal_case(source: str) -> str:
    return re.sub("[_-][a-z]", lambda mat: mat[0][1:].upper(), source.capitalize())


def param_case(source: str) -> str:
    return re.sub(
        ".[A-Z]+", lambda mat: mat[0][0] + "-" + mat[0][1:].lower(), uncapitalize(source).replace("_", "-")
    )


def snake_case(source: str) -> str:
    return re.sub(
        ".[A-Z]", lambda mat: mat[0][0] + "_" + mat[0][1:].lower(), uncapitalize(source).replace("-", "_")
    )



from .const import Empty as Empty
from .context import ContextModel as ContextModel
from .date import DateParser as DateParser
from .generic import generic_isinstance as generic_isinstance
from .generic import generic_issubclass as generic_issubclass
from .guard import is_async as is_async
from .guard import is_awaitable as is_awaitable
from .guard import is_coroutinefunction as is_coroutinefunction
from .lang import lang as lang
from .lru import LRU as LRU
from .signature import get_signature as get_signature
from .signature import signatures as signatures
from .string import split as split
from .string import split_once as split_once
from .tools import gen_subclass as gen_subclass
from .tools import group_dict as group_dict
from .tools import init_spec as init_spec
from .tools import run_always_await as run_always_await
from .tools import safe_eval as safe_eval



def split(text: str, separates: tuple[str, ...], crlf: bool = True) -> list[str]:
    """尊重引号与转义的字符串切分

    Args:
        text (str): 要切割的字符串
        separates (tuple[str, ...]): 切割符.
        crlf (bool): 是否去除 \n 与 \r，默认为 True

    Returns:
        List[str]: 切割后的字符串, 可能含有空格
    """
    ...

def split_once(text: str, separates: tuple[str, ...], crlf: bool = True) -> tuple[str, str]:
    """尊重引号与转义的字符串切分, 只切割一次

    Args:
        text (str): 要切割的字符串
        separates (tuple[str, ...]): 切割符.
        crlf (bool): 是否去除 \n 与 \r，默认为 True
    Returns:
        Tuple[str, str]: 切割后的字符串, 可能含有空格
    """
    ...



from __future__ import annotations

QUOTATION = {"'", '"'}
CRLF = {"\n", "\r"}


def split_once(text: str, separates: tuple[str, ...], crlf: bool = True):
    """尊重引号与转义的字符串切分, 只切割一次

    Args:
        text (str): 要切割的字符串
        separates (tuple[str, ...]): 切割符.
        crlf (bool): 是否去除 \n 与 \r，默认为 True

    Returns:
        Tuple[str, str]: 切割后的字符串, 可能含有空格
    """
    index, out_text, quotation, escape, sep = 0, "", "", False, False
    text = text.lstrip()
    for char in text:
        index += 1
        if (char in separates or (crlf and char in CRLF)) and not quotation:
            sep = True
            continue
        if sep:
            index -= 1
            break
        if char == "\\":
            escape = True
            out_text += char
        elif char in QUOTATION:  # 遇到引号括起来的部分跳过分隔
            if not quotation:
                quotation = char
            elif char == quotation:
                quotation = ""
            else:
                out_text += char
                continue
            if escape:
                out_text = out_text[:-1] + char
        else:
            out_text += char
            escape = False
    return out_text, text[index:]


def split(text: str, separates: tuple[str, ...], crlf: bool = True):
    """尊重引号与转义的字符串切分

    Args:
        text (str): 要切割的字符串
        separates (tuple[str, ...]): 切割符. 默认为 " ".
        crlf (bool): 是否去除 \n 与 \r，默认为 True

    Returns:
        List[str]: 切割后的字符串, 可能含有空格
    """
    result, quotation, escape = "", "", False
    for char in text:
        if char == "\\":
            escape = True
            result += char
        elif char in QUOTATION:
            if not quotation:
                quotation = char
            elif char == quotation:
                quotation = ""
            else:
                result += char
                continue
            if escape:
                result = result[:-1] + char
        elif (not quotation and char in separates) or (crlf and char in CRLF):
            if result and result[-1] != "\0":
                result += "\0"
        else:
            result += char
            escape = False
    return result.split("\0") if result else []



# cython: boundscheck=False
# cython: cdivision=True
# cython: initializedcheck=False

from cpython.list cimport PyList_Append, PyList_SET_ITEM
from cpython.object cimport Py_SIZE, PyObject
from cpython.unicode cimport PyUnicode_GET_LENGTH, PyUnicode_Join, PyUnicode_Split


cdef extern from "Python.h":
    Py_UCS4 PyUnicode_READ_CHAR(object s, Py_ssize_t i)

cdef extern from "_op.h":
    bint contains(object chs, Py_UCS4 ch)
    bint set_contains_key(object anyset, object key) except -1

cdef set QUOTES = {'"', '\''}
cdef set CRLF = {'\r', '\n'}

def split(str text, tuple separates, char crlf=True):
    """尊重引号与转义的字符串切分

    Args:
        text (str): 要切割的字符串
        separates (tuple[str, ...]): 切割符. 默认为 " ".
        crlf (bool): 是否去除 \n 与 \r，默认为 True

    Returns:
        List[str]: 切割后的字符串, 可能含有空格
    """
    cdef char escape = 0
    cdef list result = []
    cdef Py_UCS4 quotation = 0
    cdef Py_UCS4 ch = 0
    cdef Py_ssize_t i = 0
    cdef Py_ssize_t length = PyUnicode_GET_LENGTH(text)

    while i < length:
        ch = PyUnicode_READ_CHAR(text, i)
        i += 1
        if ch == 92:  # \\
            escape = 1
            PyList_Append(result, ch)
        elif set_contains_key(QUOTES, ch):
            if quotation == 0:
                quotation = ch
            elif ch == quotation:
                quotation = 0
            else:
                PyList_Append(result, ch)
                continue
            if escape:
                PyList_SET_ITEM(result, Py_SIZE(result)-1 , ch)
        elif (quotation == 0 and contains(separates, ch)) or (crlf and set_contains_key(CRLF, ch)):
            if result and result[-1] != '\1':
                PyList_Append(result, '\1')
        else:
            PyList_Append(result, ch)
            escape = 0
    if PyUnicode_GET_LENGTH(result) == 0:
        return []
    return PyUnicode_Split(PyUnicode_Join('', result), '\1', -1)


def split_once(str text, tuple separates, char crlf=True):
    text = text.lstrip()
    cdef Py_ssize_t index = 0
    cdef list out_text = []
    cdef Py_UCS4 quotation = 0
    cdef Py_UCS4 ch = 0
    cdef char escape = 0
    cdef char sep = 0
    cdef Py_ssize_t length = PyUnicode_GET_LENGTH(text)

    while index < length:
        ch = PyUnicode_READ_CHAR(text, index)
        index += 1
        if quotation == 0 and (contains(separates, ch) or (crlf and set_contains_key(CRLF, ch))):
            sep = 1
            continue
        if sep == 1:
            index -= 1
            break
        if ch == 92:  # \\
            escape = 1
            PyList_Append(out_text, ch)
        elif set_contains_key(QUOTES, ch):  # 遇到引号括起来的部分跳过分隔
            if quotation == 0:
                quotation = ch
            elif ch == quotation:
                quotation = 0
            else:
                PyList_Append(out_text, ch)
                continue
            if escape:
                PyList_SET_ITEM(out_text, Py_SIZE(out_text) - 1, ch)
        else:
            PyList_Append(out_text, ch)
            escape = 0
    return PyUnicode_Join('', out_text), text[index:]


def test_generic_isinstance():
    from typing import Any, Dict, List, Literal, TypeVar, Union

    from typing_extensions import Annotated

    from tarina import generic_isinstance

    S = TypeVar("S", bound=str)
    assert generic_isinstance("a", Any)
    assert generic_isinstance(1, int)
    assert generic_isinstance(1, Union[str, int])
    assert generic_isinstance(123, Literal[123])
    assert generic_isinstance([1], List[int])
    assert generic_isinstance(1, Annotated[int, lambda x: x > 0])
    assert generic_isinstance("a", S)
    assert generic_isinstance(1, (int, str))
    assert generic_isinstance("a", (int, str))
    assert not generic_isinstance(bool, (str, list))
    assert not generic_isinstance({123}, Dict[str, str])


def test_lru():
    """测试 LRU缓存"""
    from tarina import LRU

    cache: LRU[str, str] = LRU(3)
    cache["a"] = "a"
    cache["b"] = "b"
    cache["c"] = "c"
    assert cache.peek_first_item()[1] == "c"  # type: ignore
    _ = cache.get("a", Ellipsis)
    print(f"\n{cache.items()}")
    assert cache.peek_first_item()[1] == "a"  # type: ignore
    cache["d"] = "d"
    assert cache.get("b", Ellipsis) == Ellipsis


def test_split_once():
    """测试单次分割函数, 能以引号扩起空格, 并允许保留引号"""
    from tarina import split_once

    assert split_once("arclet-alconna", (" ",)) == ("arclet-alconna", "")
    text1 = "rrr b bbbb"
    text2 = "'rrr b' bbbb"
    text3 = "\\'rrr b\\' bbbb"
    text4 = "\\'rrr \\b\\' bbbb"
    assert split_once(text1, (" ",)) == ("rrr", "b bbbb")
    assert split_once(text2, (" ",)) == ("rrr b", "bbbb")
    assert split_once(text3, (" ",)) == ("'rrr b'", "bbbb")
    assert split_once(text4, (" ",)) == ("'rrr \\b'", "bbbb")  # 不消除其他转义字符斜杠

    assert split_once("'rrr b\" b' bbbb", (" ",)) == ('rrr b" b', "bbbb")
    assert split_once("rrr  bbbb", (" ",)) == ("rrr", "bbbb")
    assert split_once("rrr 'bbb'", (" ",)) == ("rrr", "'bbb'")


def test_split():
    """测试分割函数, 能以引号扩起空格, 并允许保留引号"""
    from tarina import split

    text1 = "rrr b bbbb"
    text2 = "'rrr b' bbbb"
    text3 = "\\'rrr b\\' bbbb"
    assert split(text1, (" ",)) == ["rrr", "b", "bbbb"]
    assert split(text2, (" ",)) == ["rrr b", "bbbb"]
    assert split(text3, (" ",)) == ["'rrr b'", "bbbb"]
    assert split("", (" ",)) == []
    assert split("  ", (" ",)) == []

    try:
        split('rrr "bbb', (" ",))
    except SyntaxError as e:
        assert str(e) == "Unterminated string: 'rrr \"bbb'"


def test_lang():
    """测试 i18n"""
    from tarina import lang
    from pathlib import Path

    assert lang.locales == {"zh-CN", "en-US"}
    lang.select("zh-CN")
    assert lang.current == "zh-CN"
    assert lang.require("lang", "error.type") == "'{target}' 在 '{locale}:{scope}' 不是合法的类型"
    assert lang.require("lang", "error.type", "en-US") == "'{target}' is not a valid type in '{locale}:{scope}'"
    lang.select("en-US")
    assert lang.current == "en-US"
    try:
        lang.select("ru-RU")
    except ValueError as e:
        assert str(e) == "'ru-RU' is not a valid language locale"
    try:
        lang.load_data("test", {})
    except KeyError as e:
        assert str(e) == "\"lang file 'test' missed require scope 'lang'\""
    lang.load_data("test", {"lang": {"error": {"type": "test"}}})
    lang.select("test")
    assert lang.require("lang", "error.type") == "test"
    assert lang.require("lang", "error.locale") == "'{target}' 不是合法的语种"
    lang.load_file(Path(__file__).parent / "en-UK.yml")
    assert lang.locales == {'zh-CN', 'test', 'en-US', 'en-UK'}


def test_init_spec():
    from dataclasses import dataclass

    from tarina import init_spec

    @dataclass
    class User:
        name: str
        age: int

    @init_spec(User)
    def add_user(user: User):
        return user

    assert add_user("test", age=1) == User("test", 1)


def test_date_parser():
    from datetime import datetime, timedelta

    from tarina import DateParser

    assert DateParser.parse("2021-01-01") == datetime(2021, 1, 1)
    assert DateParser.parse("2021-01-01 12:34:56") == datetime(2021, 1, 1, 12, 34, 56)
    assert DateParser.parse("1m").replace(microsecond=0) == datetime.now().replace(microsecond=0) + timedelta(minutes=1)
    assert DateParser.parse("h3").replace(microsecond=0) == datetime.now().replace(microsecond=0) + timedelta(hours=3)


def test_safe_eval():
    from tarina import safe_eval

    ctx = {"a": {"b": ["c"]}}

    assert safe_eval("a['b'][0]", ctx) == "c"

    class A:
        class B:
            c = 123

    assert safe_eval("A.B.c", {"A": A}) == 123



from argparse import ArgumentParser
from tarina.lang.schema import write_lang_schema
from tarina.lang.model import write_model
from pathlib import Path
import json

CONFIG_TEMPLATE = """\
{
  "default": "zh-CN",
  "frozen": [],
  "require": []
}
"""

CONFIG_INIT = """\
# This file is @generated by tarina.lang CLI tool
# It is not intended for manual editing.

from pathlib import Path

from tarina.lang import lang


lang.load(Path(__file__).parent)
"""


TEMPLATE_SCHEMA = """\
{
  "title": "Template",
  "description": "Template for lang items to generate schema for lang files",
  "type": "object",
  "properties": {
    "scopes": {
      "title": "Scopes",
      "description": "All scopes of lang items",
      "type": "array",
      "uniqueItems": true,
      "items": {
        "title": "Scope",
        "description": "First level of all lang items",
        "type": "object",
        "properties": {
          "scope": {
            "type": "string",
            "description": "Scope name"
          },
          "types": {
            "type": "array",
            "description": "All types of lang items",
            "uniqueItems": true,
            "items": {
              "oneOf": [
                {
                  "type": "string",
                  "description": "Value of lang item"
                },
                {
                  "type": "object",
                  "properties": {
                    "subtype": {
                      "type": "string",
                      "description": "Subtype name of lang item"
                    },
                    "types": {
                      "type": "array",
                      "description": "All subtypes of lang items",
                      "uniqueItems": true,
                      "items": {
                        "$ref": "#/properties/scopes/items/properties/types/items"
                      }
                    }
                  }
                }
              ]
            }
          }
        }
      }
    }
  }
}
"""

TEMPLATE_TEMPLATE = """\
{
  "$schema": ".template.schema.json",
  "scopes": []
}
"""

LANG_TEMPLATE_JSON = """\
{
  "$schema": ".lang.schema.json"
}
"""

LANG_TEMPLATE_YAML = """\
# $schema: .lang.schema.json

"""


def new(*_):
    i18n_dir = Path.cwd() / "i18n"
    if i18n_dir.exists():
        print("i18n directory already exists")
        return
    i18n_dir.mkdir()
    print(f"i18n directory created: {i18n_dir}")


def init(*_):
    root = Path.cwd()
    config_file = root / ".config.json"
    init_file = root / "__init__.py"
    template_file = root / ".template.json"
    template_schema = root / ".template.schema.json"

    with config_file.open("w+") as f:
        f.write(CONFIG_TEMPLATE)

    with init_file.open("w+") as f:
        f.write(CONFIG_INIT)

    with template_file.open("w+") as f:
        f.write(TEMPLATE_TEMPLATE)

    with template_schema.open("w+") as f:
        f.write(TEMPLATE_SCHEMA)

    print(
        f"""\
files created:
- {config_file}
- {template_file}

please edit the files to fit your needs
    """
    )


def default(args):
    root = Path.cwd()
    config_file = root / ".config.json"
    if not config_file.exists():
        print("config file not found")
        return

    with config_file.open("r") as f:
        config = json.load(f)

    if args.locale:
        config["default"] = args.locale
        with config_file.open("w") as f:
            json.dump(config, f, indent=2, ensure_ascii=False)
        print(f"default lang scope set to: {args.locale}")
    else:
        print(f"default lang scope: {config['default']}")


def schema(*_):
    root = Path.cwd()
    schema_file = root / ".lang.schema.json"
    created = not schema_file.exists()

    with (root / ".template.schema.json").open("w+") as f:
        f.write(TEMPLATE_SCHEMA)
    try:
        write_lang_schema(Path.cwd())
    except Exception as e:
        print(repr(e))
    else:
        print(f"schema for lang file {'created' if created else 'updated'}. Now you can create or update your lang files.")


def model(*_):
    root = Path.cwd()
    model_file = root / "model.py"
    init_file = root / "__init__.py"
    created = not model_file.exists()
    try:
        write_model(Path.cwd())
        with init_file.open("r") as f:
            lines = f.readlines()
        if lines[-1] != "from .model import Lang as Lang\n":
            with init_file.open("a") as f:
                f.write("\nfrom .model import Lang as Lang\n")
    except Exception as e:
        print(repr(e))
    else:
        print(f"model for lang file {'created' if created else 'updated'}. Now you can create or update your lang files.")


def create(args):
    root = Path.cwd()
    if args.yaml:
        lang_file = root / f"{args.name}.yaml"

        with lang_file.open("w+") as f:
            f.write(LANG_TEMPLATE_YAML)
    else:
        lang_file = root / f"{args.name}.json"

        with lang_file.open("w+") as f:
            f.write(LANG_TEMPLATE_JSON)

    print(f"lang file created: {lang_file}")


def delete(args):
    root = Path.cwd()
    lang_file = root / f"{args.name}.json"

    if lang_file.exists():
        lang_file.unlink()
        print(f"lang file deleted: {lang_file}")
    else:
        print(f"lang file not found: {lang_file}")


def main():
    parser = ArgumentParser(description="tarina-lang CLI tool")

    subparsers = parser.add_subparsers(dest="command")

    new_parser = subparsers.add_parser("new", help="create a new i18n directory")
    new_parser.set_defaults(func=new)

    init_parser = subparsers.add_parser("init", help="initialize a new lang configs")
    init_parser.set_defaults(func=init)

    default_parser = subparsers.add_parser("default", help="show or set default lang locale")
    default_parser.add_argument("locale", type=str, nargs="?", help="lang locale to set as default")
    default_parser.set_defaults(func=default)

    schema_parser = subparsers.add_parser("schema", help="generate or update lang schema and template schema")
    schema_parser.set_defaults(func=schema)

    model_parser = subparsers.add_parser("model", help="generate or update lang model")
    model_parser.set_defaults(func=model)

    create_parser = subparsers.add_parser("create", help="create a new lang file")
    create_parser.add_argument("name", type=str, help="name of the lang file")
    create_parser.add_argument("--yaml", action="store_true", help="create a yaml file instead of json")
    create_parser.set_defaults(func=create)

    delete_parser = subparsers.add_parser("delete", help="delete a lang file")
    delete_parser.add_argument("name", type=str, help="name of the lang file")
    delete_parser.set_defaults(func=delete)

    args = parser.parse_args()

    if args.command:
        args.func(args)
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



import json
from pathlib import Path
from typing import cast, TypedDict, Union


class _Subtypes(TypedDict):
    subtype: str
    types: list[Union[str, "_Subtypes"]]


class _TemplateDict(TypedDict):
    scope: str
    types: list[Union[str, "_Subtypes"]]


def get_template(root: Path):
    if not (root / ".template.json").exists():
        raise FileNotFoundError(f"Lang Template file not found in {root}")
    with (root / ".template.json").open("r", encoding="utf-8") as f:
        return cast(dict, json.load(f))


def schema_scope(scope: str, types: list[Union[str, "_Subtypes"]]):
    schema = {
        "title": scope.capitalize(),
        "description": f"Scope '{scope}' of lang item",
        "type": "object",
        "additionalProperties": False,
        "properties": {}
    }
    for t in types:
        if isinstance(t, str):
            schema["properties"][t] = {
                "title": t,
                "description": f"value of lang item type '{t}'",
                "type": "string"
            }
        else:
            schema["properties"][t["subtype"]] = schema_scope(t["subtype"], t["types"])
    return schema


def generate_lang_schema(root: Path):
    template = get_template(root)
    if "scopes" not in template:
        raise KeyError("Template file must have a 'scopes' key")
    scopes: list[_TemplateDict] = template["scopes"]
    return {
        "title": "Lang Schema",
        "description": f"Schema for lang file",
        "type": "object",
        "properties": {
            s["scope"]: schema_scope(s["scope"], s["types"]) for s in scopes
        },
    }


def write_lang_schema(root: Path):
    schema = generate_lang_schema(root)
    with (root / f".lang.schema.json").open("w", encoding="utf-8") as f:
        json.dump(schema, f, ensure_ascii=False, indent=2)



from __future__ import annotations

import inspect
from functools import lru_cache
from typing import Any, Awaitable, Callable, Coroutine

from typing_extensions import TypeGuard

cache_size = 4096


@lru_cache(cache_size)
def is_coroutinefunction(o) -> TypeGuard[Callable[..., Coroutine]]:
    return inspect.iscoroutinefunction(o)


@lru_cache(cache_size)
def is_awaitable(o) -> TypeGuard[Awaitable]:
    return inspect.isawaitable(o)


@lru_cache(cache_size)
def is_async(o: Any) -> TypeGuard[Callable[..., Coroutine] | Awaitable]:
    return is_coroutinefunction(o) or is_awaitable(o)



from __future__ import annotations

import contextlib
import json
import locale
import os
import sys
from pathlib import Path
from typing import Dict, Final, TypedDict, Union, cast, final

from typing_extensions import Self

root_dir: Final[Path] = Path(__file__).parent.parent / "i18n"
WINDOWS = sys.platform.startswith("win") or (sys.platform == "cli" and os.name == "nt")


class _LangDict(TypedDict):
    default: str
    frozen: list[str]
    require: list[str]


def _get_win_locale_with_ctypes() -> str | None:
    import ctypes

    kernel32 = ctypes.windll.kernel32
    lcid: int = kernel32.GetUserDefaultUILanguage()
    return locale.windows_locale.get(lcid)


def _get_win_locale_from_registry() -> str | None:
    import winreg  # noqa

    with contextlib.suppress(Exception):
        with winreg.OpenKey(winreg.HKEY_CURRENT_USER, r"Control Panel\International") as key:
            if lcid := winreg.QueryValueEx(key, "Locale")[0]:
                return locale.windows_locale.get(int(lcid, 16))


if WINDOWS:
    try:
        import ctypes

        _get_win_locale = _get_win_locale_with_ctypes
    except ImportError:
        _get_win_locale = _get_win_locale_from_registry


def get_locale() -> str | None:
    if WINDOWS:
        return _get_win_locale()

    return locale.getlocale(locale.LC_MESSAGES)[0]  # type: ignore


def _get_config(root: Path) -> _LangDict:
    if not (root / ".config.json").exists():
        raise FileNotFoundError(f"Config file not found in {root}")
    with (root / ".config.json").open("r", encoding="utf-8") as f:
        return cast(_LangDict, json.load(f))



Types = Union[str, Dict[str, "Types"]]
Raw = Dict[str, Dict[str, Types]]


def convert_dictionary(data, prefix: str = ""):
    result = {}
    for key, value in data.items():
        if isinstance(value, dict):
            result.update(convert_dictionary(value, f"{prefix}{key}."))
        else:
            result[f"{prefix}{key}"] = value
    return result


def flatten(data: Raw) -> dict[str, dict[str, str]]:
    result = {}
    for scope, types in data.items():
        result[scope] = {}
        for type, value in types.items():
            if isinstance(value, dict):
                result[scope].update(convert_dictionary({type: value}))
            else:
                result[scope][type] = value
    return result


def merge(source: dict, target: dict) -> dict:
    for key, value in source.items():
        if isinstance(value, dict):
            target[key] = merge(value, target.get(key, {}))
        elif isinstance(value, list):
            target[key] = value + target.get(key, [])
        else:
            target[key] = value
    return target


def _expand(data: list[str]):
    result: dict[str, list[str]] = {}
    for i in data:
        if "." not in i:
            i += ".*"
        parts = i.split(".", 1)
        s = parts[0]
        if s not in result:
            result[s] = []
            if not parts[1].startswith("*"):
                result[s].append(parts[1])
        elif result[s] and not parts[1].startswith("*"):
            result[s].append(parts[1])
    return result


def _get_lang(file: Path) -> Raw:
    if not file.exists():
        raise FileNotFoundError(f"Lang file '{file}' not found")
    if file.suffix.startswith(".json"):
        with file.open("r", encoding="utf-8") as f:
            data: Raw = json.load(f)
        data.pop("$schema", None)
        return data
    try:
        import yaml
    except ImportError:
        raise ImportError("PyYAML is required to load yaml file")
    with file.open("r", encoding="utf-8") as f:
        data: Raw = yaml.safe_load(f)
    data.pop("$schema", None)
    return data



def _get_scopes(root: Path) -> dict[str, dict[str, dict[str, str]]]:
    result = {}
    for i in root.iterdir():
        if not i.is_file() or i.name.startswith(".") or i.suffix not in (".json", ".yaml", ".yml"):
            continue
        result[i.stem] = flatten(_get_lang(i))
    return result


@final
class _LangConfig:
    def __init__(self):
        __config = _get_config(root_dir)
        self.__locale: str = __config["default"]
        self.__frozen: dict[str, list[str]] = _expand(__config["frozen"])
        self.__require: dict[str, list[str]] = _expand(__config["require"])
        self.__langs: dict[str, dict[str, dict[str, str]]] = _get_scopes(root_dir)
        self.__locales = set(self.__langs.keys())
        self.select_local()

    @property
    def locales(self):
        return set(sorted(self.__locales))

    @property
    def current(self):
        return self.__locale

    def select_local(self):
        """
        依据系统语言尝试自动选择语言
        """
        if (lc := get_locale()) and lc.replace("_", "-") in self.__langs:
            self.__locale = lc.replace("_", "-")
        return self

    def select(self, locale: str) -> Self:
        locale = locale.replace("_", "-")
        if locale not in self.__langs:
            raise ValueError(self.require("lang", "error.locale").format(target=locale))
        self.__locale = locale
        return self

    def save(self, root: Path | None = None):
        _root = root or root_dir
        config = _get_config(_root)
        config["default"] = self.__locale
        with (_root / ".config.json").open("w+", encoding="utf-8") as f:
            json.dump(config, f, ensure_ascii=False, indent=2)

    def load_data(self, locale: str, data: Raw):
        if locale in self.__langs:
            source = flatten(data)
            target = self.__langs[locale]
            for scope, ignores in self.__frozen.items():
                if scope not in target or scope not in source:
                    continue
                if not ignores:
                    source.pop(scope)
                else:
                    for i in ignores:
                        source[scope] = {k: v for k, v in source[scope].items() if not (k.startswith(i) and k in target[scope])}
            self.__langs[locale] = merge(source, target)
        else:
            self.__locales.add(locale)
            self.__langs[locale] = flatten(data)
        for scope, requries in self.__require.items():
            if scope not in self.__langs[locale]:
                raise KeyError(self.require("lang", "miss_require_scope", locale).format(locale=locale, target=scope))
            for t in requries:
                if any(k.startswith(t) for k in self.__langs[locale][scope]):
                    continue
                raise KeyError(self.require("lang", "miss_require_type", locale).format(locale=locale, scope=scope, target=t))

    def load_file(self, filepath: Path):
        return self.load_data(filepath.stem, _get_lang(filepath))

    def load_config(self, config: _LangDict):
        self.__locale = config.get("default", self.__locale)
        self.__frozen = merge(_expand(config.get("frozen", [])), self.__frozen)
        self.__require = merge(_expand(config.get("require", [])), self.__require)
        self.select_local()

    def load(self, root: Path) -> Self:
        self.load_config(_get_config(root))
        for i in root.iterdir():
            if not i.is_file() or i.name.startswith(".") or i.suffix not in (".json", ".yaml", ".yml"):
                continue
            self.load_file(i)
        return self

    def require(self, scope: str, type: str, locale: str | None = None) -> str:
        locale = locale or self.__locale
        if locale not in self.__langs:
            raise ValueError(self.__langs[self.__locale]["lang"]["error.locale"].format(target=locale))
        if scope in self.__langs[locale]:
            _types = self.__langs[locale][scope]
        elif scope in self.__langs[self.__locale]:
            _types = self.__langs[self.__locale][scope]
        elif scope in self.__langs[(default := _get_config(root_dir)["default"])]:
            _types = self.__langs[default][scope]
        else:
            raise ValueError(self.__langs[locale]["lang"]["error.scope"].format(target=scope, locale=locale))
        if type in _types:
            return _types[type]
        elif type in self.__langs[self.__locale][scope]:
            return self.__langs[self.__locale][scope][type]
        elif type in self.__langs[(default := _get_config(root_dir)["default"])][scope]:
            return self.__langs[default][scope][type]
        else:
            raise ValueError(self.__langs[locale]["lang"]["error.type"].format(target=type, locale=locale, scope=scope))

    def set(self, scope: str, type: str, content: str, locale: str | None = None):
        locale = locale or self.__locale
        if locale not in self.__langs:
            raise ValueError(self.__langs[self.__locale]["lang"]["error.locale"].format(target=locale))
        if scope in self.__frozen:
            raise ValueError(self.__langs[locale]["lang"]["error.scope"].format(target=scope, locale=locale))
        self.__langs[locale].setdefault(scope, {})[type] = content

    def __repr__(self):
        return f"<LangConfig: {self.__locale}>"


lang: _LangConfig = _LangConfig()

__all__ = ["lang"]



import os
import sys

from setuptools import Extension, setup

NO_EXTENSIONS = bool(os.environ.get("TARINA_NO_EXTENSIONS"))  # type: bool

if sys.implementation.name != "cpython":
    NO_EXTENSIONS = True


extensions = [
    Extension("tarina._string_c", ["src/tarina/_string_c.c"]),
    Extension("tarina._lru_c", ["src/tarina/_lru_c.c"]),
]

args = {
    "include_package_data": True,
    "exclude_package_data": {"": ["*.c"]},
}

if not NO_EXTENSIONS:
    print("**********************")
    print("* Accelerated build *")
    print("**********************")
    setup(ext_modules=extensions, **args)
else:
    print("*********************")
    print("* Pure Python build *")
    print("*********************")
    setup(**args)



from typing import Any, Callable, Generic, Hashable, Iterable, TypeVar, overload

_KT = TypeVar("_KT", bound=Hashable)
_VT = TypeVar("_VT")
_T = TypeVar("_T")

class LRU(Generic[_KT, _VT]):
    @overload
    def __init__(self, size: int, callback: None = None) -> None: ...
    @overload
    def __init__(self, size: int, callback: Callable[[_KT, _VT], Any]) -> None: ...
    def clear(self) -> None: ...
    @overload
    def get(self, key: _KT) -> _VT: ...
    @overload
    def get(self, key: _KT, instead: _VT | _T) -> _VT | _T: ...
    def get_size(self) -> int: ...
    def has_key(self, key: _KT) -> bool: ...
    def keys(self) -> list[_KT]: ...
    def values(self) -> list[_VT]: ...
    def items(self) -> list[tuple[_KT, _VT]]: ...
    def peek_first_item(self) -> tuple[_KT, _VT] | None: ...
    def peek_last_item(self) -> tuple[_KT, _VT] | None: ...
    @overload
    def pop(self, key: _KT) -> _VT: ...
    @overload
    def pop(self, key: _KT, default: _VT | _T) -> _VT | _T: ...
    def popitem(self, least_recent: bool = ...) -> tuple[_KT, _VT]: ...
    @overload
    def setdefault(self: LRU[_KT, _T | None], key: _KT) -> _T | None: ...
    @overload
    def setdefault(self, key: _KT, default: _VT) -> _VT: ...
    def set_callback(self, callback: Callable[[_KT, _VT], Any]) -> None: ...
    def set_size(self, size: int) -> None: ...
    @overload
    def update(self, __m: Iterable[tuple[_KT, _VT]], **kwargs: _VT) -> None: ...
    @overload
    def update(self, **kwargs: _VT) -> None: ...
    def __contains__(self, __o: Any) -> bool: ...
    def __delitem__(self, key: _KT) -> None: ...
    def __getitem__(self, item: _KT) -> _VT: ...
    def __len__(self) -> int: ...
    def __repr__(self) -> str: ...
    def __setitem__(self, key: _KT, value: _VT) -> None: ...



from __future__ import annotations

import sys
import types
from typing import TYPE_CHECKING, Any, List, Literal, TypeVar, Union

from typing_extensions import Annotated, get_args
from typing_extensions import get_origin as typing_ext_get_origin

if TYPE_CHECKING:
    from types import GenericAlias  # noqa
else:
    GenericAlias: type = type(List[int])

AnnotatedType: type = type(Annotated[int, lambda x: x > 0])
Unions = (Union, types.UnionType) if sys.version_info >= (3, 10) else (Union,)  # pragma: no cover


def get_origin(obj: Any) -> Any:
    return typing_ext_get_origin(obj) or obj


def generic_isinstance(obj: Any, par: type | Any | tuple[type, ...]) -> bool:
    """检查 obj 是否是 args 中的一个类型, 支持泛型, Any, Union
    Args:
        obj (Any): 要检查的对象
        par (Union[type, Any, Tuple[type, ...]]): 要检查的对象的类型
    Returns:
        bool: 是否是类型
    """
    if par is Any:
        return True
    _origin = get_origin(par)
    try:
        if isinstance(par, type):
            return isinstance(obj, par)

        if par.__class__ is AnnotatedType:
            return generic_isinstance(obj, get_args(par)[0])
        if _origin is Literal:
            return obj in get_args(par)
        if _origin in Unions:  # pragma: no cover
            for p in get_args(par):
                if generic_isinstance(obj, p):
                    return True
        if par.__class__ is tuple:
            for p in par:
                if generic_isinstance(obj, p):
                    return True
        if isinstance(par, TypeVar):  # pragma: no cover
            if par.__constraints__:
                return any(generic_isinstance(obj, p) for p in par.__constraints__)
            return generic_isinstance(obj, par.__bound__) if par.__bound__ else True
        if isinstance(obj, _origin):  # type: ignore
            return True
    except TypeError:
        return False
    return False


def generic_issubclass(cls: type, par: Union[type, Any, tuple[type, ...]]) -> bool:
    """检查 cls 是否是 args 中的一个子类, 支持泛型, Any, Union
    Args:
        cls (type): 要检查的类
        par (Union[type, Any, Tuple[type, ...]]): 要检查的类的父类
    Returns:
        bool: 是否是父类
    """
    if par is Any:
        return True
    _origin = get_origin(par)
    try:
        if isinstance(par, type):
            return issubclass(cls, par)
        if isinstance(par, AnnotatedType):
            return generic_issubclass(cls, get_args(par)[0])
        if _origin in Unions:
            for p in get_args(par):
                if generic_issubclass(cls, p):
                    return True
        if par.__class__ is tuple:
            for p in par:
                if generic_issubclass(cls, p):
                    return True
        if isinstance(par, TypeVar):
            if par.__constraints__:
                return any(generic_issubclass(cls, p) for p in par.__constraints__)
            if par.__bound__:
                return generic_issubclass(cls, par.__bound__)
        if issubclass(cls, _origin):  # type: ignore
            return True
    except TypeError:
        return False
    return False
